@@ -83,6 +83,86 @@ msgstr "bar"
         po = polib.pofile('tests/test_iso-8859-15.po', encoding='iso-8859-15')
         self.assertTrue(po.encoding == 'iso-8859-15')
 
+    def test_pofile_and_mofile8(self):
+        """
+        Test that weird occurrences are correctly parsed.
+        """
+        po = polib.pofile('tests/test_weird_occurrences.po')
+        self.assertEqual(len(po), 46)
+
+    def test_pofile_and_mofile9(self):
+        """
+        Test that obsolete previous msgid are ignored
+        """
+        po = polib.pofile('tests/test_obsolete_previousmsgid.po')
+        self.assertTrue(isinstance(po, polib.POFile))
+
+    def test_unescaped_double_quote1(self):
+        """
+        Test that polib reports an error when unescaped double quote is found.
+        """
+        data = r'''
+msgid "Some msgid with \"double\" quotes"
+msgid "Some msgstr with "double\" quotes"
+'''
+        try:
+            po = polib.pofile(data)
+            self.fail("Unescaped quote not detected")
+        except IOError:
+            exc = sys.exc_info()[1]
+            msg = 'Syntax error in po file None (line 3): unescaped double quote found'
+            self.assertEqual(str(exc), msg)
+
+    def test_unescaped_double_quote2(self):
+        """
+        Test that polib reports an error when unescaped double quote is found.
+        """
+        data = r'''
+msgid "Some msgid with \"double\" quotes"
+msgstr ""
+"Some msgstr with "double\" quotes"
+'''
+        try:
+            po = polib.pofile(data)
+            self.fail("Unescaped quote not detected")
+        except IOError:
+            exc = sys.exc_info()[1]
+            msg = 'Syntax error in po file None (line 4): unescaped double quote found'
+            self.assertEqual(str(exc), msg)
+
+    def test_unescaped_double_quote3(self):
+        """
+        Test that polib reports an error when unescaped double quote is found at the beginning of the string.
+        """
+        data = r'''
+msgid "Some msgid with \"double\" quotes"
+msgid ""Some msgstr with double\" quotes"
+'''
+        try:
+            po = polib.pofile(data)
+            self.fail("Unescaped quote not detected")
+        except IOError:
+            exc = sys.exc_info()[1]
+            msg = 'Syntax error in po file None (line 3): unescaped double quote found'
+            self.assertEqual(str(exc), msg)
+
+    def test_unescaped_double_quote4(self):
+        """
+        Test that polib reports an error when unescaped double quote is found at the beginning of the string.
+        """
+        data = r'''
+msgid "Some msgid with \"double\" quotes"
+msgstr ""
+""Some msgstr with double\" quotes"
+'''
+        try:
+            po = polib.pofile(data)
+            self.fail("Unescaped quote not detected")
+        except IOError:
+            exc = sys.exc_info()[1]
+            msg = 'Syntax error in po file None (line 4): unescaped double quote found'
+            self.assertEqual(str(exc), msg)
+    
     def test_detect_encoding1(self):
         """
         Test that given enconding is returned when file has no encoding defined.
@@ -146,6 +226,28 @@ msgstr "bar"
             polib.unescape('\\\\t and \\\\n and \\\\r and \\\\" and \\\\\\\\'),
             '\\t and \\n and \\r and \\" and \\\\'
         )
+        
+    def test_pofile_with_subclass(self):
+        """
+        Test that the pofile function correctly returns an instance of the 
+        passed in class
+        """
+        class CustomPOFile(polib.POFile):
+            pass
+        
+        pofile = polib.pofile('tests/test_indented.po', klass=CustomPOFile)
+        self.assertEqual(pofile.__class__, CustomPOFile)
+        
+    def test_mofile_with_subclass(self):
+        """
+        Test that the mofile function correctly returns an instance of the 
+        passed in class
+        """
+        class CustomMOFile(polib.MOFile):
+            pass
+        
+        mofile = polib.mofile('tests/test_utf8.mo', klass=CustomMOFile)
+        self.assertEqual(mofile.__class__, CustomMOFile)     
 
 
 class TestBaseFile(unittest.TestCase):
@@ -336,7 +438,7 @@ msgstr ""
 
 class TestPoFile(unittest.TestCase):
     """
-    Tests for MoFile class.
+    Tests for PoFile class.
     """
 
     def test_save_as_mofile(self):
@@ -394,6 +496,10 @@ class TestPoFile(unittest.TestCase):
         po = polib.pofile('tests/test_pofile_helpers.po')
         self.assertEqual(len(po.obsolete_entries()), 4)
 
+    def test_unusual_metadata_location(self):
+        po = polib.pofile('tests/test_unusual_metadata_location.po')
+        self.assertNotEqual(po.metadata, {})
+        self.assertEqual(po.metadata['Content-Type'], 'text/plain; charset=UTF-8')
 
 class TestMoFile(unittest.TestCase):
     """
